/************************************************************************************************
* Vessel Parameter MFD -- displays various vessel parameters									*
*																								*
* Usage:																						*
*		Vessel Parameters MFD is a multi-functional display mode for the Orbiter simulator,		*
*		version 2006-P1, by Dr. Martin Schweiger. Its primary purpose is to provide various		*
*		information about the user's current focus spacecraft.									*
*																								*
* Parameters:																					*
*		Name: Name of the vessel.																*
*		Class: Class of the vessel.																*
*																								*
*		Mass: This will have a toggle to switch between mass in kilograms and metric tons		*
*			-Dry: Total mass without propellant in kilograms or metric tons.					*
*			-Wet: Total mass with all propellant tanks fully filled in kilograms or metric tons.*
*			-Current: Total mass of the vessel at the present time in kilograms or metric tons.	*
*																								*
*		Tanks: This will have a toggle to switch to each tank sequentially.						*
*			-Full: Maximum capacity of the tank in kilograms.									*
*			-Current: Present mass of propellant in the tank, as either a value in kilograms	* 
*					  or a percentage of total capacity. This will have a toggle to switch		*
*					  between the two modes.													*
*			-Engines serviced: Which engines or thrusters use this tank for propellant. All		*
*					  engines serviced are displayed at one time.								*
*																								*
*		Engines and Thrusters: There will be a toggle to switch between engines sequentially.	*
*			-Thrust: Nominal thrust at full throttle, in kilonewtons.							*
*			-Specific Impulse: This is the vacuum ISP of the engine, given in N*s/kg.			*
*			-Acceleration:																		*
*			--Linear: Current acceleration in m/s^2 and the axis relative to the vessel that	*
*					  it accelerates along, again assuming nominal full thrust.					*
*			--Rotational: Current acceleration in degrees/s^2 or radians/s^2, again assuming	*
*						  nominal full thrust. This also provides the direction of the			*
*						  acceleration.															*
*			-Tanks: Which tanks the engine draws propellant from.								*
*																								*
*		Attachments:																			*
*			-Name: Name of the attachment.														*
*			-Mass: Mass of the attachment.														*
*																								*
* Restrictions:																					*
*		All of the figures given assume that the vessel is undocked with no additional mass		*
*		beyond that specified in either its vessel .dll or spacecraft configuration file. The	*
*		accuracy of this data becomes unreliable if	either assumption is not valid.				*
*																								*
* Limitations:																					*
*		Staging cannot be taken into account due to the way that Orbiter handles staged vessels.*
*		Presumably the wet and dry weights correspond to whether the tanks of the current stage	*
*		are full or empty respectively, but this is uncertain. While the mass of attachments	*
*		is considered, their effect on performance is not as Orbiter does not natively consider	*
*		them for performance.																	*
*																								*
* Change Log:																					*
*		11 August 2009: Added SI notation and made it the default. Release Cantidate 3			*
*		9 August 2009: Fixed a few minor bugs. Release cantidate 2.								*
*		7 August 2009: Completed Engines display mode. Release cantidate 1.						*
*		31 July 2009: Completed Tanks display mode.												*
*		27 July 2009: Compiled first version. Fixed a pair of obvious bugs.						*
*		19 July 2009: Changed Full and Empty to Wet and Dry respectively. Added limitations on `*
*		staging and	attachments.																*
*		19 July 2009: First draft of preliminary specification drawn up.						*
*************************************************************************************************/

/*
	Copyright 2009 by Blake Chapman (nemoricus@gmail.com)

	This file is part of Vessel Parameters MFD.

	Vessel Parameters MFD is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

	doubleToSI and numDecimals from [http://www.orbithangar.com/searchid.php?ID=3609 FreeOrbitMFD by CJ Plooy] were 
	modified and used for formatting the SI prefixes.

*/

#include "VesselParametersMFD.h"
#include <math.h>

// =====================================================================
// VesselParametersMfd functions
// =====================================================================

// *********************************************************************
// Constructor for the Vessel Parameters MFD .
// *********************************************************************

VesselParametersMfd::VesselParametersMfd(UINT mfd, DWORD w, DWORD h, VESSEL *v): MFD (w, h, v)
{
	vpMfdInfo = vesselParametersMfdManager->registerMfd(this, v, mfd); //This also gets a relevant VesselParametersMfdInfo instance.
}

// *********************************************************************
// Destructor for the Vessel Parameters MFD .
// *********************************************************************

VesselParametersMfd::~VesselParametersMfd()
{
	vesselParametersMfdManager->unRegisterMfd(this);
}

// *********************************************************************
// GetWetMass
//
// This gets the total mass of the current vessel. This function exists
// to make up for the fact that Orbiter lacks a method to do this natively.
// *********************************************************************

double VesselParametersMfd::GetWetMass(double dry)
{
	DWORD index; //Holds the current number tank.
	DWORD MAX_INDEX; //Holds the total number of propellant resources.
	double total_mass; //Holds the total mass of the vessel.

	MAX_INDEX = pV->GetPropellantCount(); //Storing the number of tanks the vessel has.
	total_mass = dry; //The dry mass is the baseline mass that the rest is added to.

	for(index = 0; index < MAX_INDEX; index++)
	{
		total_mass += pV->GetPropellantMaxMass(pV->GetPropellantHandleByIndex(index)); 
		//This adds the mass of each propellant tank to the baseline mass.
	}

	return total_mass;
}

// *********************************************************************
// GetDigits - Returns the number of digits in a number.
// *********************************************************************

inline int GetDigits(int number)
{
	int digits = 1;

	// Each division by ten removes a place from the number. 
	// The number of divisions to decrease the value to less than ten is 
	// the number of digits in the number.
	while(number > 10) 
	{
		number /= 10;
		++digits;
	}

	return digits;
}

void VesselParametersMfd::DisplayThrustersServiced(HDC hDC)
{
	DWORD index; //Controls the for loop.
	DWORD MAX_INDEX; //This will hold the number of thrusters the vessel has.
	THRUSTER_HANDLE th; //This is a holding variable for the thruster handle.
	PROPELLANT_HANDLE ph; //This a holding variable for the propellant tank handle.
	char thrusters_serviced[sizeOfBuffer]; //This stores the output string.
	int stored_characters = 0; //This holds the number of characters after the initial part currently stored in the string.
	int character_increment = 2; //This is the value by which stored_characters is incremented.
	int output_line = 9; //This variable is used to control the output line if more than one is needed.
	double max_isp = 0; //Holds the highest ISP of the thrusters connected to the tank.
	double thruster_isp; //Holds the ISP of the current thruster.

	MAX_INDEX = pV->GetThrusterCount(); //Storing the total number of thrusters that the vessel has.
	
	//The handle for the propellant resource only needs to be gotten once, 
	//so it can be done outside the loop.
	ph = pV->GetPropellantHandleByIndex(vpMfdInfo->tank_index);

	sprintf_s(thrusters_serviced, sizeOfBuffer, "Thrusters Serviced:");

	for(index = 0; index < MAX_INDEX; ++index)
	{
		th = pV->GetThrusterHandleByIndex(index);

		if(pV->GetThrusterResource(th) == ph) //If the thruster's propellant resource is the propellant tank being displayed...
		{
			//Record the number of the thruster.
			sprintf_s(thrusters_serviced, sizeOfBuffer, "%s %d", thrusters_serviced, (index + 1));

			/* Ensure that the character increment is the right size.
			The value sent to get digits is incremented by one as the displayed value counts 
			from one while the internal index counts from zero.
			The returned value is also increased for one to account for the space between 
			the previous and the current number. */
			character_increment = GetDigits(index + 1) + 1; 
			stored_characters += character_increment;

			thruster_isp = pV->GetThrusterIsp0(th);

			if(thruster_isp > max_isp) max_isp = thruster_isp;
		}

		if((stored_characters + character_increment) >= 17) //The field that the numbers are displayed in is only seventeen characters wide.
		{
			TextXY(hDC, 0, output_line, thrusters_serviced);
			++output_line; //If a line is printed, we need to move to the next one.
			strncpy_s(thrusters_serviced, "                   ", sizeOfBuffer); //Some whitespace to keep alignment.
			stored_characters = 0;
		}
	}

	if(stored_characters > 0)//If there are still characters in the buffer, they need to be printed.
	{
		TextXY(hDC, 0, output_line, thrusters_serviced);
	}

	//If the output line is still 8 and the number of stored characters is zero, then there are
	//no thrusters serviced by this propellant resource.
	if(output_line == 9 && stored_characters == 0)
	{
		sprintf_s(thrusters_serviced, sizeOfBuffer, "%s %s", thrusters_serviced, "None");
		TextXY(hDC, 0, 9, thrusters_serviced);
		TextXY(hDC, 0, 8, "Delta-V: None");
	}

	else
	{
		double deltav;
		double vessel_mass;
		double tank_mass;
		char output_buffer[sizeOfBuffer];

		vessel_mass = pV->GetMass();
		tank_mass = pV->GetPropellantMass(ph);

		deltav = max_isp * log(vessel_mass / (vessel_mass - tank_mass)); //Tsiolkovsky's rocket equation.
		
		formatUnsignedNumber(output_buffer, deltav);
		TextXY(hDC, 0, 8, "Delta-V: %sm/s", output_buffer);
	}
}

// *********************************************************************
// formatUnsignedNumber
//
// Prepares numbers without an explicit sign for output.
// *********************************************************************

void VesselParametersMfd::formatUnsignedNumber(char *result, double value)
{
	switch(vpMfdInfo->notation)
	{
		case SI:
			unsignedDoubleToSI(result, value);
			break;
		case EXP:
			sprintf_s(result, sizeOfBuffer, "%11.3e ", value);
			break;
	}
		return;
}

// *********************************************************************
// formatSignedNumber
//
// Prepares numbers with an explicit sign for output.
// *********************************************************************

void VesselParametersMfd::formatSignedNumber(char *result, double value)
{
	switch(vpMfdInfo->notation)
	{
		case SI:
			signedDoubleToSI(result, value);
			break;
		case EXP:
			sprintf_s(result, sizeOfBuffer, "%+11.3e ", value);
			break;
	}
		return;
}

// *********************************************************************
// numDecimals - returns the number of decimals required for outputting the
//   values to four significant digits
// *********************************************************************
int VesselParametersMfd::numDecimals(double value)
{
	value = fabs(value);

	int ret = 0;
	while(value < 1000.0 && ret < 3)
	{
		value *= 10.0;
		++ret;
	}
	return ret;
}

// *********************************************************************
// unsignedDoubleToSI - converts a double that is not explicitly signed to a 
//   string containing the number formatted using SI units to 4 
//   significant digits
// *********************************************************************
void VesselParametersMfd::unsignedDoubleToSI(char *result, double value)
{
	double a = fabs(value);

	if(a == 0)
	{
		sprintf_s(result, sizeOfBuffer, "%0.*f ", numDecimals(value), value);
	}

	else if(a < 1e-6)      //Less than one one-millionth.
	{
		value /= 1e-9;
		sprintf_s(result, sizeOfBuffer, "%0.*f n", numDecimals(value), value);
	}

	else if(a < 1e-3)      //Less than one one-thousanth, greater than one one-millionth.
	{
		value /= 1e-6;
		sprintf_s(result, sizeOfBuffer, "%0.*f ", numDecimals(value), value);
	}

	else if(a < 1e0)      //Less than one, greater than one one-thousanth.
	{
		value /= 1e-3;
		sprintf_s(result, sizeOfBuffer, "%0.*f m", numDecimals(value), value);
	}

	else if(a < 1e3)      //Less than one thousand.
	{
		sprintf_s(result, sizeOfBuffer, "%0.*f ", numDecimals(value), value);
	}
	else if(a < 1e6)    //Greater than one thousand, less than one million.
	{
		value /= 1e3;
		sprintf_s(result, sizeOfBuffer, "%0.*f k", numDecimals(value), value);
	}
	else if(a < 1e9)    //Greater than one million, less than one billion.
	{
		value /= 1e6;
		sprintf_s(result, sizeOfBuffer, "%0.*f M", numDecimals(value), value);
	}
	else if(a < 1e12)    //Greater than one billion, less than one trillion.
	{
		value /= 1e9;
		sprintf_s(result, sizeOfBuffer, "%0.*f G", numDecimals(value), value);
	}
	else if(a < 1e15)    //Greater than one trillion, less than one quadrillion.
	{
		value /= 1e12;
		sprintf_s(result, sizeOfBuffer, "%0.*f T", numDecimals(value), value);
	}
	else if(a < 1e18)    //Greater than one quadrillion, less than one quintillion.
	{
		value /= 1e15;
		sprintf_s(result, sizeOfBuffer, "%0.*f P", numDecimals(value), value);
	}
}

// *********************************************************************
// unsignedDoubleToSI - converts a double that is explicitly signed to a 
//   string containing the number formatted using SI units to 4 
//   significant digits
// *********************************************************************
void VesselParametersMfd::signedDoubleToSI(char *result, double value)
{
	double a = fabs(value);

	if(a == 0)
	{
		sprintf_s(result, sizeOfBuffer, "%+0.*f ", numDecimals(value), value);
	}

	else if(a < 1e-6)      //Less than one one-millionth.
	{
		value /= 1e-9;
		sprintf_s(result, sizeOfBuffer, "%+0.*f n", numDecimals(value), value);
	}

	else if(a < 1e-3)      //Less than one one-thousanth, greater than one one-millionth.
	{
		value /= 1e-6;
		sprintf_s(result, sizeOfBuffer, "%+0.*f ", numDecimals(value), value);
	}

	else if(a < 1e0)      //Less than one, greater than one one-thousanth.
	{
		value /= 1e-3;
		sprintf_s(result, sizeOfBuffer, "%+0.*f m", numDecimals(value), value);
	}

	else if(a < 1e3)      //Less than one thousand.
	{
		sprintf_s(result, sizeOfBuffer, "%+0.*f ", numDecimals(value), value);
	}
	else if(a < 1e6)    //Greater than one thousand, less than one million.
	{
		value /= 1e3;
		sprintf_s(result, sizeOfBuffer, "%+0.*f k", numDecimals(value), value);
	}
	else if(a < 1e9)    //Greater than one million, less than one billion.
	{
		value /= 1e6;
		sprintf_s(result, sizeOfBuffer, "%+0.*f M", numDecimals(value), value);
	}
	else if(a < 1e12)    //Greater than one billion, less than one trillion.
	{
		value /= 1e9;
		sprintf_s(result, sizeOfBuffer, "%+0.*f G", numDecimals(value), value);
	}
	else if(a < 1e15)    //Greater than one trillion, less than one quadrillion.
	{
		value /= 1e12;
		sprintf_s(result, sizeOfBuffer, "%+0.*f T", numDecimals(value), value);
	}
	else if(a < 1e18)    //Greater than one quadrillion, less than one quintillion.
	{
		value /= 1e15;
		sprintf_s(result, sizeOfBuffer, "%+0.*f P", numDecimals(value), value);
	}
}

// *********************************************************************
// Updates the Display for the MFD.
//
// This function uses MFDlib to ease putting information on screen.
// *********************************************************************

void VesselParametersMfd::Update (HDC hDC)
{
	InvalidateButtons(); //Just to make things look smooth if the buttons have changed.
	// initialise MFDlib
	MFDUpdate(hDC,cw,ch);

	char title[sizeOfBuffer]; //Will hold the title of the MFD.
	char mode[sizeOfBuffer]; //Will hold the MFD Mode.
	char vessel_name[sizeOfBuffer]; //Will hold the name of the vessel.
	char class_name[sizeOfBuffer]; //Will hold the class name of the vessel.

	switch(vpMfdInfo->parametersMode) //This selects the appropriate mode to be displayed immediately after the title.
	{
		default:
			vpMfdInfo->parametersMode = MASS;
			//Fall through.
		case MASS:
			strcpy_s(mode, "Mass");
			break;

		case TANKS:
			strcpy_s(mode, "Tanks");
			break;

		case ENGINES:
			strcpy_s(mode, "Engines");
			break;
	}

	sprintf_s(title, sizeOfBuffer, "VesselParametersMFD: %s", mode);
	Title(hDC, title); //This prints out the title of our MFD, which in this case also includes its current mode.

	//Getting and outputting data.
	SetMFDColor(hDC, WHITE); //Setting the color of all following output text to white.
	sprintf_s(vessel_name, sizeOfBuffer, "%s", pV->GetName()); //Storing the name of the vessel.
	sprintf_s(class_name, sizeOfBuffer, "%s", pV->GetClassName()); //Storing the class of the vessel.
	
	TextXY(hDC, 0, 2, "Name: %s", vessel_name); //Outputting the vessel name.
	TextXY(hDC, 0, 3, "Class: %s", class_name); //Outputting the vessel class.

	switch(vpMfdInfo->parametersMode) //This selects the appropriate output mode.
	{
		case MASS:
			DisplayMass(hDC);
			break;

		case TANKS:
			DisplayTanks(hDC);
			break;

		case ENGINES:
			DisplayEngines(hDC);
			break;
	}
}

// *********************************************************************
// DisplayMass - Displays the data for mass.
// *********************************************************************

void VesselParametersMfd::DisplayMass(HDC hDC)
{
	char dry_mass[sizeOfBuffer]; //Holds the fully empty mass of the vessel.
	char wet_mass[sizeOfBuffer]; //Holds the fully loaded mass of the vessel.
	char current_mass[sizeOfBuffer]; //Holds the current mass of the vessel.
	char units[sizeOfBuffer]; //Holds the units for certain outputs.
	struct
	{
		double dry;
		double wet;
		double current;
	} mass_data;

	mass_data.dry = pV->GetEmptyMass(); //Storing the empty mass of the vessel.
	mass_data.wet = GetWetMass(mass_data.dry); //Storing the wet mass of the vessel.
	mass_data.current = pV->GetMass(); //Storing the current mass of the vessel.
	if(vpMfdInfo->massdisplay == METRIC_TONS)
	{
		//This converts the mass into metric tons.
		mass_data.dry /= 1000;
		mass_data.wet /= 1000;
		mass_data.current /= 1000;
		strcpy_s(units, sizeOfBuffer, "t");
	}
	else 
	{
		if(vpMfdInfo->notation == SI)
		{
			//This converts the mass to grams so that it's output with the right prefix.
			mass_data.dry *= 1000;
			mass_data.wet *= 1000;
			mass_data.current *= 1000;
			strcpy_s(units, sizeOfBuffer, "g");
		}

		else strcpy_s(units, sizeOfBuffer, "kg"); //Otherwise I just need kilograms.
	}

	formatUnsignedNumber(dry_mass, mass_data.dry); //This puts the numerical data into a string format that can be printed.
	TextXY(hDC, 0, 5, "Dry Mass:     %s%s", dry_mass, units); //This outputs the dry mass.
	formatUnsignedNumber(wet_mass, mass_data.wet); //This and the next three lines do the same for wet and current mass respectivey.
	TextXY(hDC, 0, 6, "Wet Mass:     %s%s", wet_mass, units);
	formatUnsignedNumber(current_mass, mass_data.current);
	TextXY(hDC, 0, 7, "Current Mass: %s%s", current_mass, units);
}

// *********************************************************************
// DisplayTanks - Displays the data for tanks.
// *********************************************************************

void VesselParametersMfd::DisplayTanks(HDC hDC)
{
	char full_mass[sizeOfBuffer]; //This will hold the string for the full mass of the tank.
	char current_mass[sizeOfBuffer]; //This will hold the string for the current mass in the tank.
	char thrusters_serviced[sizeOfBuffer] = "";
	char units[sizeOfBuffer]; //Holds the units for certain outputs.
	struct
	{
		double full; //This will hold the value for the mass in the full tank.
		double current; //This will hold the current value for the mass in the tank.
	} tank_data;

	if(vpMfdInfo->tank_index >= pV->GetPropellantCount())
		vpMfdInfo->tank_index = 0; //Sanity check. If a tank disappears, I don't want to try and reference it.

	//These get and store the maximum and current values of the mass of propellant in the tank.
	tank_data.full = pV->GetPropellantMaxMass(pV->GetPropellantHandleByIndex(vpMfdInfo->tank_index));
	tank_data.current = pV->GetPropellantMass(pV->GetPropellantHandleByIndex(vpMfdInfo->tank_index));

	if(vpMfdInfo->notation == SI)
	{
		//This converts the mass to grams so that it's output with the right prefix.
		tank_data.full *= 1000;
		tank_data.current *= 1000;
		strcpy_s(units, sizeOfBuffer, "g");
	}

	else strcpy_s(units, sizeOfBuffer, "kg"); //Otherwise I just need kilograms.

	TextXY(hDC, 0, 5, "Tank %d", (vpMfdInfo->tank_index + 1)); //The plus one switches from zero based counting to one based counting.
	formatUnsignedNumber(full_mass, tank_data.full); //This prepares the tank data for output and outputs it.
	TextXY(hDC, 0, 6, "Full Mass: %s%s", full_mass, units);

	if(vpMfdInfo->massdisplay == PERCENTAGE)
	{
		if(tank_data.full != 0) //Sanity check. Divide by zero errors are messy.
		{
			tank_data.current = (tank_data.current / tank_data.full) * 100; //Switching the data from kilograms to percentage.
			TextXY(hDC, 0, 7, "Current Mass: %8.3lf %%", tank_data.current);
		}
		
		else TextXY(hDC, 0, 7, "Divide by zero error.");
	}

	else 
	{
		formatUnsignedNumber(current_mass, tank_data.current);
		TextXY(hDC, 0, 7, "Current Mass: %s%s", current_mass, units);
	}

	DisplayThrustersServiced(hDC); //Displaying the thrusters that the tank services.
}

// *********************************************************************
// DisplayEngines - Displays the data for engines and thrusters.
// *********************************************************************

void VesselParametersMfd::DisplayEngines(HDC hDC)
{
	char output_buffer[sizeOfBuffer]; //This holds the data after it has been formatted for output.
	char units[sizeOfBuffer]; //Holds the units for certain outputs.
	THRUSTER_HANDLE th; //This holds the handle for the thruster.
	VECTOR3 total_thrust; //This holds the vector of the thruster.
	VECTOR3 position; //This holds the position of the thruster.
	VECTOR3 direction; //This holds the direction of the thrust vector.
	VECTOR3 radial_thrust; //This will hold the radial thrust generated by the thruster.
	VECTOR3 torque; //This will hold the torque the thruster generates.
	VECTOR3 pmi; //This will hold the vessel's principal moments of inertia.
	double vessel_mass; //This stores the vessel's current mass for use in acceleration calculations.

	struct
	{
		double vacuum_thrust; //Thrust of the engine in kilonewtons.
		double isp; //ISP of the engine in N*s/kg
		double linear_x; //Linear acceleration components of the engine.
		double linear_y;
		double linear_z;
		double angular_x; //Angular acceleration components of the engine.
		double angular_y;
		double angular_z;
	} engine_data;

	if(vpMfdInfo->engine_index >= pV->GetThrusterCount())
		vpMfdInfo->engine_index = 0; //Sanity check. If an engine disappears, I don't want to try and reference it.

	th = pV->GetThrusterHandleByIndex(vpMfdInfo->engine_index);
	vessel_mass = pV->GetMass();
	pV->GetPMI(pmi);
	engine_data.vacuum_thrust = pV->GetThrusterMax0(th); //This gets the thrust in Newtons.
	engine_data.isp = pV->GetThrusterIsp0(th); //This gets the vacuum ISP in N*s/kg.
	pV->GetThrusterRef(th, position); //This gets the position that the thruster thrusts from.
	pV->GetThrusterDir(th, direction); //This gets the direction that the thruster thrusts in.

	total_thrust = direction; //This gives the lengths of the thrust direction vector's components.
	total_thrust /= length(total_thrust); //This converts the thrust direction vector into a unit vector.
	total_thrust *= engine_data.vacuum_thrust; //This converts the unit vector into the thrust vector.

	//This determines the thrust along each axis.
	if(length(position) == 0) //This covers the event that the position vector is the origin.
	{
		radial_thrust = total_thrust;	
		torque.x = 0;
		torque.y = 0;
		torque.z = 0;
	}

	else
	{
		radial_thrust = (direction / length(direction)) * -(dotp(total_thrust, (position/length(position))));
		torque = -crossp(total_thrust, position); //This determines the torque around each axis.
	}

	engine_data.linear_x =  radial_thrust.x / vessel_mass; //This determines the acceleration in m/s^2 along the x axis.
	engine_data.linear_y =  radial_thrust.y / vessel_mass;
	engine_data.linear_z =  radial_thrust.z / vessel_mass;

	engine_data.angular_x = torque.x / (pmi.x * vessel_mass); //This determines the angular acceleration around each axis.
	engine_data.angular_y = torque.y / (pmi.y * vessel_mass);
	engine_data.angular_z = torque.z / (pmi.z * vessel_mass);

	if(vpMfdInfo->notation == SI)
	{
		strcpy_s(units, sizeOfBuffer, "N");
	}
	else
	{
		engine_data.vacuum_thrust /= 1000;
		strcpy_s(units, sizeOfBuffer, "kN");
	}

	//Outputting data.
	TextXY(hDC, 0, 5, "Engine %d", vpMfdInfo->engine_index + 1); //Outputting engine number.
	formatUnsignedNumber(output_buffer, engine_data.vacuum_thrust);
	TextXY(hDC, 0, 6, "Vacuum Thrust: %s%s", output_buffer, units); //Outputting vacuum thrust in kN.
	formatUnsignedNumber(output_buffer, engine_data.isp);	
	TextXY(hDC, 0, 7, "ISP: %sN*s/kg", output_buffer); //ISP.
	TextXY(hDC, 0, 8, "Acceleration:");
	TextXY(hDC, 0, 9, "    Linear:");
	formatSignedNumber(output_buffer, engine_data.linear_x);
	TextXY(hDC, 0, 10, "        X: %sm/s^2", output_buffer); //Linear acceleration components.
	formatSignedNumber(output_buffer, engine_data.linear_y);
	TextXY(hDC, 0, 11, "        Y: %sm/s^2", output_buffer);
	formatSignedNumber(output_buffer, engine_data.linear_z);
	TextXY(hDC, 0, 12, "        Z: %sm/s^2", output_buffer);

	if(vpMfdInfo->enginesdisplay == DEGREES)
	{
		engine_data.angular_x *= 180 / PI;
		engine_data.angular_y *= 180 / PI;
		engine_data.angular_z *= 180 / PI;
		strcpy_s(units, sizeOfBuffer, "deg/s^2");
	}

	else strcpy_s(units, sizeOfBuffer, "rad/s^2");

	TextXY(hDC, 0, 13, "    Angular:");
	formatSignedNumber(output_buffer, engine_data.angular_x);
	TextXY(hDC, 0, 14, "        X: %s%s", output_buffer, units);
	formatSignedNumber(output_buffer, engine_data.angular_y);
	TextXY(hDC, 0, 15, "        Y: %s%s", output_buffer, units);
	formatSignedNumber(output_buffer, engine_data.angular_z);
	TextXY(hDC, 0, 16, "        Z: %s%s", output_buffer, units);

	PROPELLANT_HANDLE ph = pV->GetThrusterResource(th);

	if(ph == NULL)
	{
		TextXY(hDC, 0, 17, "Propellant Resource: Not Connected");
	}

	else
	{
		DWORD index = 0;
		DWORD MAX_INDEX  = pV->GetPropellantCount();

		for(index = 0; index < MAX_INDEX; ++index)
		{
			if(pV->GetPropellantHandleByIndex(index) == ph)
				break;
		}

		TextXY(hDC, 0, 17, "Propellant Resource: Tank %d", index + 1);
	}
}

// *********************************************************************
// ButtonLabel - returns the button labels.
// *********************************************************************

char *VesselParametersMfd::ButtonLabel(int bt)
{
	switch(bt)
	{
		case 0:
			return "MOD";
		case 1:
			return "UNT";
		case 2:
			return "NOT";
		case 3:
			if(vpMfdInfo->parametersMode == TANKS) //This button label isn't needed in any other modes.
				return "TNK";
			else if(vpMfdInfo->parametersMode == ENGINES)
				return "ENG";
			else return NULL;
		default:
			return NULL;
	}
}

// *********************************************************************
// ButtonMenu - Returns the help text.
// *********************************************************************

int VesselParametersMfd::ButtonMenu(const MFDBUTTONMENU **menu) const
{
	const static MFDBUTTONMENU mnu[4] = 
	{
		{"Change mode", 0, 0},
		{"Select units", 0, 0},
		{"Changes notation", 0, 0},
		{"Changes tank/engine", 0, 0},
	};

	if (menu) *menu = mnu;//I have no idea what this does.
	return numCmds;
}

// *********************************************************************
// ConsumeButton - Called to handle mouse clicks.
// *********************************************************************

bool VesselParametersMfd::ConsumeButton(int bt, int event)
{
	if(!(event & PANEL_MOUSE_LBDOWN)) return false; // We're only interested in left mouse clicks.

	switch (bt)
	{
		case 0:
			if(vpMfdInfo->parametersMode == MASS)
			{
				vpMfdInfo->parametersMode = TANKS;
				if(vpMfdInfo->massdisplay == METRIC_TONS) vpMfdInfo->massdisplay = KILOGRAMS;
			}
			else if(vpMfdInfo->parametersMode == TANKS)
			{
				vpMfdInfo->parametersMode = ENGINES;
			}
			else
			{
				vpMfdInfo->parametersMode = MASS;
				if(vpMfdInfo->massdisplay != METRIC_TONS || vpMfdInfo->massdisplay != KILOGRAMS) 
					vpMfdInfo->massdisplay = KILOGRAMS;
			}
			return true;

		case 1: // UNT
			switch(vpMfdInfo->parametersMode)
			{
				case MASS:
					if(vpMfdInfo->massdisplay == KILOGRAMS)
						vpMfdInfo->massdisplay = METRIC_TONS;
					else
						vpMfdInfo->massdisplay = KILOGRAMS;
					break;

				case TANKS:
					if(vpMfdInfo->massdisplay == KILOGRAMS)
						vpMfdInfo->massdisplay = PERCENTAGE;
					else
						vpMfdInfo->massdisplay = KILOGRAMS;
					break;

				case ENGINES:
					if(vpMfdInfo->enginesdisplay == RADIANS)
						vpMfdInfo->enginesdisplay = DEGREES;
					else
						vpMfdInfo->enginesdisplay = RADIANS;
					break;

			}
			return true;

		case 2:
			if(vpMfdInfo->notation == EXP)
				vpMfdInfo->notation = SI;
			else
				vpMfdInfo->notation = EXP;
			return true;

		case 3:
			if(vpMfdInfo->parametersMode == TANKS)
			{
				++(vpMfdInfo->tank_index);
				if(vpMfdInfo->tank_index >= pV->GetPropellantCount())
					vpMfdInfo->tank_index = 0;
			}

			else if(vpMfdInfo->parametersMode == ENGINES)
			{
				++(vpMfdInfo->engine_index);
				if(vpMfdInfo->engine_index >= pV->GetThrusterCount())
					vpMfdInfo->engine_index = 0;
			}
			return true;
	}

	return false;
}

// *********************************************************************
// WriteStatus - Writes information to the scenario file.
// *********************************************************************
void VesselParametersMfd::WriteStatus(FILEHANDLE scn) const
{
	oapiWriteScenario_int(scn,"MFD_ID", vpMfdInfo->mfdId);
	switch(vpMfdInfo->parametersMode)
	{
		default: //Fall through.
		case MASS:
			oapiWriteScenario_string(scn,"PARAMETERSMODE", "MASS");
			break;
		case TANKS:
			oapiWriteScenario_string(scn,"PARAMETERSMODE", "TANKS");
			break;
		case ENGINES:
			oapiWriteScenario_string(scn,"PARAMETERSMODE", "ENGINES");
			break;
	}

	oapiWriteScenario_int(scn,"TANK_INDEX", vpMfdInfo->tank_index);
	oapiWriteScenario_int(scn,"ENGINE_INDEX", vpMfdInfo->engine_index);

	if(vpMfdInfo->notation == SI)
		oapiWriteScenario_string(scn,"NOTATION", "SI");
	else
		oapiWriteScenario_string(scn,"NOTATION", "EXP");

	if(vpMfdInfo->massdisplay == METRIC_TONS)
		oapiWriteScenario_string(scn,"MASSDISPLAY", "METRIC_TONS");
	else if(vpMfdInfo->massdisplay == PERCENTAGE)
		oapiWriteScenario_string(scn,"MASSDISPLAY", "PERCENTAGE");
	else
		oapiWriteScenario_string(scn,"MASSDISPLAY", "KILOGRAMS");

	if(vpMfdInfo->enginesdisplay == DEGREES)
		oapiWriteScenario_string(scn,"ENGINESDISPLAY", "DEGREES");
	else
		oapiWriteScenario_string(scn,"ENGINESDISPLAY", "RADIANS");
}

// *********************************************************************
// ReadStatus - Loads information from the scenario file.
// *********************************************************************

void VesselParametersMfd::ReadStatus(FILEHANDLE scn)
{
	char *line;

	while(oapiReadScenario_nextline(scn,line))
	{
		if(!_strnicmp(line, "MFD_ID", 6))
		{
			if(!_strnicmp(line+7, "0", 1))
			{
				if(vpMfdInfo->mfdId == 0)
					break;
			}

			else if(!_strnicmp(line+7, "1", 1))
			{
				if(vpMfdInfo->mfdId == 1)
					break;
			}
		}
	}

	while (oapiReadScenario_nextline(scn,line))
	{
		if ( !_strnicmp(line, "PARAMETERSMODE", 14) )
		{
			if ( !_strnicmp(line+15, "TANKS", 5) )
				vpMfdInfo->parametersMode = TANKS;

			else if( !_strnicmp(line+15, "ENGINES", 7) )
				vpMfdInfo->parametersMode = ENGINES;
			else
				vpMfdInfo->parametersMode = MASS;
		}

		else if ( !_strnicmp(line,"TANK_INDEX", 10) )
		{
			sscanf_s(line+11, "%d", &(vpMfdInfo->tank_index));
		}

		else if ( !_strnicmp(line,"ENGINE_INDEX", 12) )
		{
			sscanf_s(line+13, "%d", &(vpMfdInfo->engine_index));
		}

		if ( !_strnicmp(line, "NOTATION", 8) )
		{
			if ( !_strnicmp(line+9, "EXP", 3) )
				vpMfdInfo->notation = EXP;
			else
				vpMfdInfo->notation = SI;
		}

		else if ( !_strnicmp(line,"MASSDISPLAY", 11) )
		{
			if ( !_strnicmp(line+12, "METRIC_TONS", 11) )
				vpMfdInfo->massdisplay = METRIC_TONS;
			else if ( !_strnicmp(line+12, "PERCENTAGE", 10) )
				vpMfdInfo->massdisplay = PERCENTAGE;
			else
				vpMfdInfo->massdisplay = KILOGRAMS;
		}

		else if ( !_strnicmp(line,"ENGINESDISPLAY", 13) )
		{
			if ( !_strnicmp(line+14, "DEGREES", 7) )
				vpMfdInfo->enginesdisplay = DEGREES;
			else
				vpMfdInfo->enginesdisplay = RADIANS;
		}

		else if ( !_strnicmp(line,"END_MFD", 7) )
			break;
	}
}

// =====================================================================
// VesselParametersMfdInfo functions
// =====================================================================

// *********************************************************************
// Constructor
// *********************************************************************

VesselParametersMfdInfo::VesselParametersMfdInfo(VesselParametersMfd *hMfd, VESSEL *v, UINT mfd)
{
	vessel = v;
	mfdId = mfd;
	vpMfd = hMfd;
	tank_index = 0;
	engine_index = 0;
	massdisplay = KILOGRAMS;
	enginesdisplay = RADIANS;
	notation = SI;
	parametersMode = MASS;
}

// =====================================================================
// VesselParametersMfdManager functions
// =====================================================================

// *********************************************************************
// Constructor for the Vessel Parameters MFD manager.
// *********************************************************************

VesselParametersMfdManager::VesselParametersMfdManager(HINSTANCE hModule)
{
	//This is used by Orbiter to manage certain aspects of the MFD.
	MFDMODESPEC spec;

	// Hold onto the handle for later use.
	hDll = hModule;

	//Record specific aspects of the MFD.
	spec.key = NULL; //This used to set the key used to select the MFD mode. However, it is no longer used.
					 //Still, it is safe to set it to null to make sure that nothing weird happens.
	spec.name = "Vessel Parameters"; //Should be self-explanatory. This is the name of the MFD as displayed to the user.
	spec.msgproc = msgProc; //Evidently this is the address of MFD message parser. However, I'm not sure what that means.
	
	mfdMode = oapiRegisterMFDMode(spec);
}

// *********************************************************************
// Destructor for the Vessel Parameters MFD manager.
// *********************************************************************

VesselParametersMfdManager::~VesselParametersMfdManager()
{
	std::vector<VesselParametersMfdInfo*>::iterator i;

	for (i = vesselParametersMfdInfos.begin(); i < vesselParametersMfdInfos.end(); i++)
	{
		delete (*i);
	}
	
	oapiUnregisterMFDMode(mfdMode); //This tells Orbiter that your MFD is no longer in use and to free up the resources
									//allocated to it.
}

// *********************************************************************
// msgProc - Processes messages from Orbiter
// *********************************************************************
int VesselParametersMfdManager::msgProc(UINT msg, UINT mfd, WPARAM wparam, LPARAM lparam)
{
	if (msg == OAPI_MSG_MFD_OPENED)
		return (int)(new VesselParametersMfd(mfd,LOWORD(wparam),HIWORD(wparam),(VESSEL*)lparam));

	return 0;
}

// *********************************************************************
// saveState - save state to scenario file
// *********************************************************************
void VesselParametersMfdManager::saveState(FILEHANDLE scn)
{
	std::vector<VesselParametersMfdInfo*>::iterator i;

	for (i = vesselParametersMfdInfos.begin(); i != vesselParametersMfdInfos.end(); i++)
	{
		if((*i)->vpMfd == NULL) //The active MFD instances have stored their own status.
		{
			oapiWriteScenario_string(scn,"BEGIN_VP_MFD_INFO","");

			oapiWriteScenario_string(scn,"  VESSEL",(*i)->vessel->GetName());
			oapiWriteScenario_int(scn,"TANK_INDEX", (*i)->tank_index);
			oapiWriteScenario_int(scn,"ENGINE_INDEX", (*i)->engine_index);
			oapiWriteScenario_int(scn, "  MFD_ID", (*i)->mfdId);

			if((*i)->notation == SI)
				oapiWriteScenario_string(scn,"NOTATION", "SI");
			else
				oapiWriteScenario_string(scn,"NOTATION", "EXP");

			//Record the Vessel Parameters display mode.
			switch ( (*i)->parametersMode )
			{
				default: //Fall through
				case MASS:
					oapiWriteScenario_string(scn,"  PARAMETERSMODE", "MASS");
					break;
				case TANKS:
					oapiWriteScenario_string(scn,"  PARAMETERSMODE", "TANKS");
					break;
				case ENGINES:
					oapiWriteScenario_string(scn,"  PARAMETERSMODE", "ENGINES");
					break;
			}

			//Record the mass display mode.
			switch ( (*i)->massdisplay )
			{
				default: //Fall through
				case KILOGRAMS:
					oapiWriteScenario_string(scn,"  MASSDISPLAY", "KILOGRAMS");
					break;
				case PERCENTAGE:
					oapiWriteScenario_string(scn,"  MASSDISPLAY", "PERCENTAGE");
					break;
				case METRIC_TONS:
					oapiWriteScenario_string(scn,"  MASSDISPLAY", "METRIC_TONS");
					break;
			}

			//Record the mass display mode.
			switch ( (*i)->enginesdisplay )
			{
				default: //Fall through
				case RADIANS:
					oapiWriteScenario_string(scn,"  ENGINESDISPLAY", "RADIANS");
					break;
				case DEGREES:
					oapiWriteScenario_string(scn,"  ENGINESDISPLAY", "DEGREES");
					break;
			}

			oapiWriteScenario_string(scn,"END_VP_MFD_INFO","");
		}
	}
}


// *********************************************************************
// loadState - save state to scenario file
// *********************************************************************

void VesselParametersMfdManager::loadState(FILEHANDLE scn)
{
	char *line;
	char cbuf[sizeOfBuffer];
	VesselParametersMfdInfo *vpibuf;

	while (oapiReadScenario_nextline(scn,line))
	{
		if ( !_strnicmp(line,"BEGIN_VP_MFD_INFO",17) )
		{
			vpibuf = new VesselParametersMfdInfo(NULL, NULL, 0);
			
			while (oapiReadScenario_nextline(scn,line))
			{
				if ( !_strnicmp(line,"VESSEL",6) )
				{
					sscanf_s(line+7,"%s", cbuf, sizeOfBuffer);
					if ( !(vpibuf->vessel = oapiGetVesselInterface(oapiGetVesselByName(cbuf))) )
						break; // go no further if the vessel is not valid
				}

				else if ( !_strnicmp(line,"TANK_INDEX", 10) )
				{
					sscanf_s(line+11, "%d", &(vpibuf->tank_index));
				}

				else if ( !_strnicmp(line,"ENGINE_INDEX", 12) )
				{
					sscanf_s(line+13, "%d", &(vpibuf->engine_index));
				}

				else if ( !_strnicmp(line, "MFD_ID", 6) )
				{
					sscanf_s(line+7,"%d",&(vpibuf->mfdId));
				}

				if ( !_strnicmp(line, "NOTATION", 8) )
				{
					if ( !_strnicmp(line+9, "EXP", 3) )
						vpibuf->notation = EXP;
					else
						vpibuf->notation = SI;
				}

				else if ( !_strnicmp(line, "PARAMETERSMODE", 14) )
				{
					if ( !_strnicmp(line+15, "TANKS", 5) )
						vpibuf->parametersMode = TANKS;
					else if ( !_strnicmp(line+15, "ENGINES", 7) )
						vpibuf->parametersMode = ENGINES;
					else
						vpibuf->parametersMode = MASS; // defaults to MASS if not TANKS or ENGINES
				}

				else if ( !_strnicmp(line, "MASSDISPLAY", 11) )
				{
					if ( !_strnicmp(line+12, "METRIC_TONS", 12) )
						vpibuf->massdisplay = METRIC_TONS;
					else if ( !_strnicmp(line+12, "PERCENTAGE", 10) )
						vpibuf->massdisplay = PERCENTAGE;
					else
						vpibuf->massdisplay = KILOGRAMS; // defaults to KILOGRAMS
				}

				else if ( !_strnicmp(line, "ENGINESDISPLAY", 14) )
				{
					if ( !_strnicmp(line+15, "DEGREES", 7) )
						vpibuf->enginesdisplay = DEGREES;
					else
						vpibuf->enginesdisplay = RADIANS; // defaults to RADIANS
				}

				else if ( !_strnicmp(line,"END_VP_MFD_INFO",15) )
					break;
			}

			if ((vpibuf->vessel)) // check info has valid handles
				vesselParametersMfdManager->vesselParametersMfdInfos.push_back(vpibuf); // add it to the list if OK
			else
				delete vpibuf; // discard it otherwise

			if ( (!_strnicmp(line,"END",3)) && (strlen(line) == 3) )
				break; // stop reading if we've reached the end of the block
		}
	}
}

// *********************************************************************
// registerMfd - register an MFD instance
// *********************************************************************

VesselParametersMfdInfo* VesselParametersMfdManager::registerMfd(VesselParametersMfd *hMfd, VESSEL* v, UINT mfd) 
{
	std::vector<VesselParametersMfdInfo*> mfdInfoBuf; //This buffer holds the relevant VesselParametersMfdInfo instances.
	std::vector<VesselParametersMfdInfo*>::iterator i;

	vesselParametersMfds.push_back(hMfd); //This stores the handle for the MFD.

	//This makes a list of all of the VesselParametersMfdInfo that are relevant to the desired vessel.
	for (i = vesselParametersMfdInfos.begin();i != vesselParametersMfdInfos.end(); i++)
	{
		if ( (*i)->vessel == v )
			mfdInfoBuf.push_back(*i); 
	}

	//If there are no instances of the VesselParametersMfdInfo class that point to that vessel, make one that does.
	if ( mfdInfoBuf.size() == 0 )
	{
		vesselParametersMfdInfos.push_back(new VesselParametersMfdInfo(hMfd,v,mfd));
		return vesselParametersMfdInfos.back();
	}

	//If there are, find the one that is in that MFD location.
	for (i = mfdInfoBuf.begin(); i != mfdInfoBuf.end(); i++)
	{
		if ( (*i)->mfdId == mfd )
		{
			(*i)->vpMfd = hMfd;
			return (*i);
		}
	}

	//Failing that, find a currently unused one and apply to the current MFD.
	for (i = mfdInfoBuf.begin(); i != mfdInfoBuf.end(); i++)
	{
		if ( (*i)->vpMfd == NULL ) 
		{
			(*i)->vpMfd = hMfd;
			return (*i);
		}
	}

	vesselParametersMfdInfos.push_back(new VesselParametersMfdInfo(hMfd,v,mfd)); // use new info if none other suitable found
	return vesselParametersMfdInfos.back();
}

// *********************************************************************
// unRegisterMfd - unregister an MFD instance
// *********************************************************************

void VesselParametersMfdManager::unRegisterMfd(VesselParametersMfd *hMfd) // unregisters an MFD instance
{
	std::vector<VesselParametersMfd*>::iterator i;

	for (i = vesselParametersMfds.begin(); i != vesselParametersMfds.end(); i++)
	{
		if ( (*i) == hMfd )
		{
			(*i)->vpMfdInfo->vpMfd = NULL;
			i=vesselParametersMfds.erase(i);
			return;
		}
	}
}

void VesselParametersMfdManager::deleteVessel(VESSEL *v)
{
	std::vector<VesselParametersMfdInfo*>::iterator i;

	for (i = vesselParametersMfdInfos.begin(); i < vesselParametersMfdInfos.end(); i++)
	{
		if ( (*i)->vessel == v ) i = vesselParametersMfdInfos.erase(i);
	}
}
// =====================================================================
// Orbiter API Interface
// =====================================================================


//This function is called by Orbiter to initialize your addon.
DLLCLBK void InitModule(HINSTANCE hModule)
{
	if (!vesselParametersMfdManager) //You don't want to create a new instance of the manager if one already exists.
		vesselParametersMfdManager = new VesselParametersMfdManager(hModule);
}

//This function is called by Orbiter to remove your addon.
DLLCLBK void ExitModule(HINSTANCE hModule)
{
	delete vesselParametersMfdManager;
	vesselParametersMfdManager = NULL; //DON'T LEAVE DANGLING POINTERS! They cause bad things to happen.
									  //Nulling this prevents that.
}

// -------------------------------------------------------------------------
// opcSaveState - save the state to scenario file
// -------------------------------------------------------------------------
DLLCLBK void opcSaveState(FILEHANDLE scn)
{
	vesselParametersMfdManager->saveState(scn);
}

// -------------------------------------------------------------------------
// opcLoadState - load the state to scenario file
// -------------------------------------------------------------------------
DLLCLBK void opcLoadState(FILEHANDLE scn)
{
	vesselParametersMfdManager->loadState(scn);
}

// -------------------------------------------------------------------------
// opcDeleteVessel - use to purge vessel from any cache before deletion
// -------------------------------------------------------------------------
DLLCLBK void opcDeleteVessel(OBJHANDLE hVessel)
{
	VESSEL *v;

	v = oapiGetVesselInterface(hVessel);
	vesselParametersMfdManager->deleteVessel(v);
}