/*
	Copyright 2009 by Blake Chapman (nemoricus@gmail.com)

	This file is part of Vessel Parameters MFD.

	Vessel Parameters MFD is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

	doubleToSI and numDecimals from [http://www.orbithangar.com/searchid.php?ID=3609 FreeOrbitMFD by CJ Plooy] were 
	modified and used for formatting the SI prefixes.

*/

#ifndef VESSELPARAMETERS_MFD_H
#define VESSELPARAMETERS_MFD_H

#define ORBITER_MODULE

#include <vector>
#include "Orbitersdk.h"
#include "MFDlib.h"
#include "VesselParametersMFDTypeDefs.h"

// *************************************************************************
// Global variables
// *************************************************************************

//Handle for MFD Manager. 
//Used to send information to and from the manager.
//It's intialized to null to keep weird things from happening.
static VesselParametersMfdManager *vesselParametersMfdManager = NULL;
static const int sizeOfBuffer = 64; //A size value to use in any character arrays.

// *************************************************************************
// VesselParametersMfd class
//
// Implements Vessel Parameters MFD, i.e., this is what gets information to
// the user.
// *************************************************************************

class VesselParametersMfd:public MFD
{
private:
	void formatUnsignedNumber(char *result, double value);
	void formatSignedNumber(char *result, double value);
	void unsignedDoubleToSI(char *result, double value);
	void signedDoubleToSI(char *result, double value);
	int numDecimals(double value);
	double GetWetMass(double dry);
	void DisplayMass(HDC hDC);
	void DisplayTanks(HDC hDC);
	void DisplayEngines(HDC hDC);
	void DisplayThrustersServiced(HDC hDC);

	VesselParametersMfdInfo* vpMfdInfo;
	static const int numCmds = 4;
public:
	VesselParametersMfd(UINT mfd, DWORD w, DWORD h, VESSEL *v); //Constructor
	~VesselParametersMfd(); //Destructor

	//Functions overwritten from the MFD base class.
	void Update (HDC hDC); //Updates the display. Must be overwritten as the base function is purely virtual.
	char *ButtonLabel(int bt); //Returns the labels for the MFD buttons.
	int ButtonMenu(const MFDBUTTONMENU **menu) const; //This function is called to provide the help text 
													  //when the MNU button is pressed.
	bool ConsumeButton(int bt, int event); //This is called to handle mouse clicks on MFD buttons.
	void WriteStatus(FILEHANDLE scn) const; //Called when it's time to write data to the scenario file.
	void ReadStatus(FILEHANDLE scn); //Called when it's time to load data from the scenario file.

	friend VesselParametersMfdManager;
	friend VesselParametersMfdInfo;
};

// *************************************************************************
// VesselParametersMfdInfo class
//
// Stores the settings for inactive and active MFD instances.
// *************************************************************************

class VesselParametersMfdInfo
{
private:
	ParametersMode parametersMode;
	Massdisplay massdisplay;
	Enginesdisplay enginesdisplay;
	Notation notation;
	DWORD tank_index;
	DWORD engine_index;
	VESSEL *vessel;
	VesselParametersMfd *vpMfd;
	UINT mfdId;

public:
	VesselParametersMfdInfo(VesselParametersMfd *hMfd, VESSEL *v, UINT mfd); //Constructor
	//~VesselParametersMfdInfo(); Default destructor.

	friend VesselParametersMfd;
	friend VesselParametersMfdManager;
};

// *************************************************************************
// VesselParametersMfdManager class
//
// Manages all active VesselParametersMfd.
// *************************************************************************

class VesselParametersMfdManager
{
private:
	VesselParametersMfdInfo* registerMfd(VesselParametersMfd *hMfd, VESSEL* v, UINT mfd); // register an MFD instance
	void unRegisterMfd(VesselParametersMfd *hMfd); // unregisters an MFD instance.

	HINSTANCE hDll; // Handle to the module dll.
	int mfdMode; // Stores MFD mode.
	std::vector<VesselParametersMfd*> vesselParametersMfds; //Stores the location of all VesselParametersMfd in memory.
	std::vector<VesselParametersMfdInfo*> vesselParametersMfdInfos; //Stores the location of all VesselParametersMfdInfo in memory.

public:
	VesselParametersMfdManager(HINSTANCE hModule); //Constructor
	~VesselParametersMfdManager(); //Destructor

	void saveState(FILEHANDLE scn); //Hanldes loadingo of inative MFDs.
	void loadState(FILEHANDLE scn); //Handles saving of inactive MFDs.
	void deleteVessel(VESSEL *v); //This removes any reference to a vessel that's been deleted.

	static int msgProc(UINT msg, UINT mfd, WPARAM wparam, LPARAM lparam); // Used to handle messages from Orbiter.

	friend VesselParametersMfd;
};

// *************************************************************************
// Orbiter API interface
// *************************************************************************

// -------------------------------------------------------------------------
// InitModule - called when module initially loaded
// -------------------------------------------------------------------------
DLLCLBK void InitModule(HINSTANCE hModule);

// -------------------------------------------------------------------------
// ExitModule - called just before module is unloaded
// -------------------------------------------------------------------------
DLLCLBK void ExitModule(HINSTANCE hModule);

// -------------------------------------------------------------------------
// opcSaveState - save the state to scenario file
// -------------------------------------------------------------------------
DLLCLBK void opcSaveState(FILEHANDLE scn);

// -------------------------------------------------------------------------
// opcLoadState - load the state to scenario file
// -------------------------------------------------------------------------
DLLCLBK void opcLoadState(FILEHANDLE scn);

// -------------------------------------------------------------------------
// opcDeleteVessel - use to purge vessel from any cache before deletion
// -------------------------------------------------------------------------
DLLCLBK void opcDeleteVessel(OBJHANDLE hVessel);

#endif //!VESSELPARAMETERS_MFD_H